/* Copyright (C) 1989 Aladdin Enterprises.  All rights reserved.
   Distributed by Free Software Foundation, Inc.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* gxdraw.c */
/* Primitive drawing routines for GhostScript imaging library */
#include "gx.h"
#include "gxfixed.h"
#include "gxmatrix.h"
#include "gzcolor.h"
#include "gzdevice.h"
#include "gzstate.h"

gx_bitmap *gx_tile_expand(P3(int, int, gs_state *));

/* These routines all assume that gx_color_render has been called. */

/* Fill a rectangle. */
int
gz_fill_rectangle(int x, int y, int w, int h, gs_color *pcolor, gs_state *pgs)
{	int darker = pcolor->dev_color.color1;
	int lighter;
	gx_device *dev = pgs->device->info;
	gx_bitmap *tile;
	int code;
#ifdef gs_DEBUG
if ( gs_debug['q'] )
	printf("[q]x=%d y=%d w=%d h=%d  c1=%d c2=%d htl=%d\n",
		x, y, w, h, darker, pcolor->dev_color.color2,
		pcolor->dev_color.halftone_level);
#endif
	if ( color_is_pure(pcolor) )	/* no halftoning */
	   {	(*dev->procs->fill_rectangle)(dev, x, y, w, h, darker);
		return 0;
	   }
	lighter = pcolor->dev_color.color2;
	tile = pcolor->dev_color.tile;
	/* Try to tile the rectangle primitively */
	code = (*dev->procs->tile_rectangle)(dev, tile, x, y, w, h, darker, lighter);
	if ( code < 0 )
	   {	/* Use the default implementation */
		code = gx_default_tile_rectangle(dev, tile, x, y, w, h, darker, lighter);
	   }
	return code;
}

/* Fill a trapezoid.  This routine is not satisfactory, */
/* because it doesn't handle fractional coordinates properly. */
/* (The interface will have to be changed eventually.) */
/* Requirements: wt >= 0, wb >= 0. */
int
gz_fill_trapezoid(int xt, int wt, int yt, int xb, int wb, int h,
  gs_color *pcolor, gs_state *pgs)
{	gx_device *dev = pgs->device->info;
	if ( wt == 0 && wb == 0 || h <= 0 ) return 0;
	if ( color_is_pure(pcolor) &&
	    (*dev->procs->fill_trapezoid)(dev, xt, yt, wt, xb, yt + h, wb,
					  pcolor->dev_color.color1) >= 0 )
	  return 0;
	   {	int x, fx;
		int dx, dx1, dxf = xb - xt;
		int w, fw;
		int dw, dw1, dwf = wb - wt;
		int y, yb;
		int rx, ry, rw;
		int abs_dx;
		/* Compute integer and fractional deltas */
#define reduce_delta(df, d, d1, pos)\
	if ( df >= 0 )\
	   {	if ( df >= h )\
		  d1 = (d = df / h) + 1, df -= d * h;\
		else	/* save the divides */\
		   {	pos();\
			d = 0, d1 = 1;\
		   }\
	   }\
	else			/* df < 0 */\
	   {	if ( df <= -h )\
		  d1 = (d = df / h) - 1, df = d * h - df;\
		else	/* save the divides */\
		  d = 0, d1 = -1, df = -df;\
	   }
#define pos_for_x()\
	if ( dxf == 0 && dwf == 0 )	/* detect rectangle */\
	   {	gz_fill_rectangle(xt, yt, wt, h, pcolor, pgs);\
		return 0;\
	   }
		reduce_delta(dxf, dx, dx1, pos_for_x);
		/* If the sides are very shallow, intermediate points won't */
		/* be drawn.  Correct for this now. */
		abs_dx = abs(dx);
		if ( abs_dx > 1 )
		   {	/* Increase the width */
			if ( wt < abs_dx ) wt = abs_dx;
			if ( wb < abs_dx ) wb = abs_dx;
			dwf = wb - wt;
		   }
#define pos_for_w()			/* nothing */
		reduce_delta(dwf, dw, dw1, pos_for_w);
		x = xt, fx = dxf >> 1;
		w = wt, fw = 0;
		y = yt, yb = y + h;
		rx = x, ry = y, rw = w;
		/* Do the fill */
		do
		   {	if ( x != rx || w != rw )	/* detect rectangles */
			   {	gz_fill_rectangle(rx, ry, rw, y - ry, pcolor, pgs);
				rx = x, rw = w, ry = y;
			   }
			if ( (fx += dxf) >= h ) fx -= h, x += dx1;
			else x += dx;
			if ( (fw += dwf) >= h ) fw -= h, w += dw1;
			else w += dw;
		   }
		while ( ++y < yb );
		if ( y != ry )
		   {	gz_fill_rectangle(rx, ry, rw, y - ry, pcolor, pgs);
		   }
	   }
	return 0;
}

/* Default implementation of tile_rectangle */
int
gx_default_tile_rectangle(gx_device *dev, register gx_bitmap *tile,
  int x, int y, int w, int h, int color0, int color1)
{	/* Fill the rectangle in chunks */
	int width = tile->width;
	int raster = tile->raster;
	int ex = x + w, ey = y + h;
	int fex = ex - width, fey = ey - tile->height;
	int irx = ((width & (width - 1)) == 0 ?	/* power of 2 */
		x & (width - 1) :
		x % width);
	int ry = y % tile->height;
	int icw = width - irx;
	int ch = tile->height - ry;
	byte *row = tile->data + ry * raster;
	int (*proc)(P10(gx_device *, byte *, int, int, int, int, int, int, int, int)) = dev->procs->copy_mono;
	int cx, cy;
#ifdef gs_DEBUG
if ( gs_debug['t'] )
   {	int ptx, pty;
	byte *ptp = tile->data;
	printf("[t]tile w=%d h=%d raster=%d; x=%d y=%d w=%d h=%d\n",
		tile->width, tile->height, tile->raster, x, y, w, h);
	for ( pty = 0; pty < tile->height; pty++ )
	   {	printf("   ");
		for ( ptx = 0; ptx < tile->raster; ptx++ )
			printf("%3x", *ptp++);
	   }
	printf("\n");
   }
#endif
	if ( icw > w ) icw = w;
	if ( ch > h ) ch = h;
#define real_copy_tile(sourcex, tx, tw)\
  (*proc)(dev, row, sourcex, raster, tx, cy, tw, ch, color0, color1)
#ifdef gs_DEBUG
#define copy_tile(sx, tx, tw)\
  if ( gs_debug['t'] )\
	printf("   copy sx=%d x=%d y=%d w=%d h=%d\n",\
		sx, tx, cy, tw, ch);\
  real_copy_tile(sx, tx, tw)
#else
#define copy_tile(sx, tx, tw)\
  real_copy_tile(sx, tx, tw)
#endif
	for ( cy = y; cy < ey; )
	   {	copy_tile(irx, x, icw);
		cx = x + icw;
		while ( cx <= fex )
		   {	copy_tile(0, cx, width);
			cx += width;
		   }
		if ( cx < ex )
		   {	copy_tile(0, cx, ex - cx);
		   }
		cy += ch;
		ch = (cy > fey ? ey - cy : tile->height);
		row = tile->data;
	   }
#undef copy_tile
#undef real_copy_tile
	return 0;
}

/* Draw a one-pixel-wide line. */
int
gz_draw_line(int ix, int iy, int itox, int itoy,
  gs_color *pcolor, gs_state *pgs)
{	if ( iy == itoy )	/* horizontal line */
	   {	int w = itox - ix + 1;
		if ( w <= 0 )
			gz_fill_rectangle(itox, iy, 2 - w, 1, pcolor, pgs);
		else
			gz_fill_rectangle(ix, iy, w, 1, pcolor, pgs);
	   }
	else
	   {	int h = itoy - iy + 1;
		gx_device *dev = pgs->device->info;
		if ( h <= 0 )
		   {	/* Fill primitive requires h > 0 */
			int tix = ix;
			ix = itox, iy = itoy;
			itox = tix;
			h = 2 - h;
		   }
		if ( color_is_pure(pcolor) &&
		    (*dev->procs->draw_line)(dev, ix, iy, itox, iy + h,
					     pcolor->dev_color.color1) >= 0 )
		  return 0;
		gz_fill_trapezoid(ix, 1, iy, itox, 1, h, pcolor, pgs);
	   }
	return 0;
}

/****** STUBS ******/
int gx_default_draw_line() { return -1; }
int gx_default_fill_trapezoid() { return -1; }
int gx_default_tile_trapezoid() { return -1; }
